/* Copyright (C) Krzysztof Strasburger, 2009
   licencja GNU GPL, wersja 2 lub pniejsza */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "mw3.h"
#include "wektor.h"
#include "disp.h"

/* Struktura "wektor" jest wykorzystywana troche inaczej,
   niz w obliczeniach rozchodzenia sie swiatla;
   elementy "omin" i "hmin" sluza do przechowywania przesuniecia opisu
   wzgledem pozycji domyslnej */

extern void blad();
extern char *xmalloc();
extern void macierz_obrotu();
extern struct dispfn * disp_init();

struct dispfn *dfn = (struct dispfn *) NULL;

char seplist[] = " \t\r\n";
int dshow = 1;

static double rzut(double *v, double *v0, double rr0) {
	double c;
	c = adotb3(v, v0);
	if (c <= 0.)	/* nie trafia w ekran */
		return c;
	c = rr0 / c;	/* skalowanie wektora, by jego koniec lezal na ekranie */
	dscal3(v, c);
	aminusb3(v, v0, v);	/* W v otrzymujemy wektor na plaszczyznie ekranu */
	return sqrt(adotb3(v,v));	/* Zwracana jest dlugosc tego wektora */
}

/* Obliczenie wspolrzednych punktow na ekranie:
   wl - lista struktur zawierajacych wektory kierunkowe,
   wc - wektor prowadzacy do srodka ekranu, wp - wektor pionu,
   d - odleglosc od punktu obserwacji do srodka ekranu */
static void przelicz(struct wektor *wl, double *wc, double *wp, double d) {
	double c, rr, d1, d2, sa, ca, w1[3], w2[3], p[3];
	/* wektor wyznaczajacy srodek ekranu */
	c = d / sqrt(adotb3(wc,wc));	/* wspolczynnik skalujacy */
	dscal3(wc, c);
	rr = d * d;
	/* wektor wyznaczajacy dodatnia polos y (pion) */
	dcopy3(wp, w1);
	d1 = rzut(w1, wc, rr);
	/* w1 powinien zawierac rzut wektora pionu na ekran */
	if (d1 <= 1.e-10*d)	/* nie trafia w ekran */
		blad("Wektor wyznaczajacy pion nie trafia w ekran lub jest zbyt krotki.");
	/* Petla po elementach listy */
	while (wl) {
		if (wl->vis < 0)	/* Element umyslnie pominiety */
			goto koniec;
		dcopy3(wl->w, w2);
		d2 = rzut(w2, wc, rr);
		if (d2 <= 0.) {	/* nie trafia w ekran */
			wl->vis = 0;
			printf("Obiekt %s jest poza ekranem.\n",wl->nazwa);
			goto koniec;
		} else {
			wl->vis = 1;
		}
		/* d2 - wspolrzedna biegunowa r; znajdujemy kosinus i sinus
		   kata miedzy rzutami pionu i wektora biezacego na ekran */
		ca = adotb3(w2, w1) / (d1 * d2);
		sa = 1. - ca * ca;
		/* zabezpieczenie przed niestabilnosciami numerycznymi */
		if (sa <= 0.)
			sa = 0.;
		else
			sa = sqrt(sa);
		/* Znak sinusa! Iloczyn wektorowy w1 i w ma kierunek
		   taki, jak wektor wyznaczajacy srodek ekranu, */
		axb3(w1, w2, p);
		/* ale jego zwrot moze byc przeciwny. */
		if (adotb3(wc, p) > 0.)
			sa = -sa;
		/* Obliczamy wspolrzedne kartezjanskie na ekranie. */
		/* Musimy pamietac, ze wektor w1 wyznacza pion, a nie poziom. */
		wl->y = d2 * ca;
		wl->x = -d2 * sa;
koniec:
		wl=wl->n;
	}
}

/* Biezace informacje o punktach na ekranie */
void pokaz(FILE *f, struct wektor *wb) {
	while (wb) {
		if (wb->vis > 0) {
			fprintf(f,"%.5lf %.5lf %.3lf %.1lf %.1lf %.1lf %s",wb->x,wb->y,wb->o,wb->h,wb->omin,wb->hmin,wb->nazwa);
			if (dshow)
				fprintf(f,"(%.1lf)",wb->o);
			fprintf(f,"\n");
		}
		wb = wb->n;
	}
}

/* static inline void piszw(char *c, double *a) {
	printf("%s: %lg %lg %lg l=%lg\n",c,a[0],a[1],a[2],sqrt(adotb3(a,a)));
} */

/* Poszukiwanie obiektu o danej nazwie na liscie */
struct wektor *scan_list(struct wektor *wl, char *nazwa_s) {
	while (wl) {
		if (strcmp(wl->nazwa, nazwa_s) == 0)
			return wl;
		wl = wl->n;
	}
	printf("Nie znaleziono obiektu %s\n",nazwa_s);
	return (struct wektor *) NULL;
}

/* Ustawienie znacznika widzialnosci na obiektach z listy */
void vislist(struct wektor *w, int val) {
	struct wektor *wb;
	char *c;
	while ((c = strtok((char *) NULL, seplist)) && c) {
		wb = scan_list(w, c);
		if (wb) {
			if (wb->vis) {
				wb->vis = val;
				printf("Ustawiono znacznik widzialnosci obiektu %s.\n",wb->nazwa);
			} else
				printf("Obiekt %s jest poza polem widzenia.\n",wb->nazwa);
		}
	}
}

/* Funkcja zwracajaca liczbe podwojnej precyzji z kolejnego parametru polecenia */
double dblcmd() {
	double d;
	char *par;
	par = strtok((char *) NULL, seplist);
	if (par)
		d = atof(par);
	else
		d = 0.;
	return d;
}

#define CMDL 256
/* Petla interpretera; polecenia:
   rc a - obrot o kat a wokol wektora wskazujacego srodek ekranu,
   rv a - obrot o kat a wokol wektora pionowego w punkcie obserwacji
	  (zmiana kierunku obserwacji),
   rh a - obrot o kat a wokol wektora poziomego w punkcie obserwacji
	  (pochylenie aparatu fotograficznego),
   d odleglosc - ustawienie odleglosci (w metrach) ekranu od obserwatora,
   snazwa_pliku - zapisanie biezacego stanu do pliku,
   hnazwa_obiektu - schowanie obiektu o danej nazwie,
   unazwa_obiektu - pokazanie obiektu o danej nazwie,
   lnazwa_obiektu h - przesuniecie podpisu obiektu o h,
   wc x y z - wczytanie wektora wskazujacego srodek ekranu,
   wp x y z - wczytanie wektora wskazujacego pion,
   p - pokazanie punktow
   c komenda - komenda przekazana do programu roboczego
   k - koniec obliczen.
*/
static void interp(struct wektor *w, double *c, double *p) {
	double d, pp[3], ph[3], m[9], t[3], a;
	double *w1, *w2, *w3;
	struct wektor *wb;
	char cmd[CMDL];
	char *scmd, *par;
	int n;
	FILE *f, *fi;
	/* Inicjujemy odleglosc od ekranu */
	d = 1000.;
	/* Znajdujemy pomocnicze wektory: poziomu (pochylanie aparatu) */
	axb3(c, p, ph);
	/* i pionu (obrot aparatu wokol osi pionowej) */
	axb3(ph, c, pp);
	fi = stdin;
	do {
powtorz:
		printf("ekran> ");
		fflush(stdout);
		if (fgets(cmd, CMDL, fi) != cmd) {
			if (fi == stdin) {
				perror("fgets");
				blad("blad odczytu polecenia");
			}
			fclose(fi);
			fi = stdin;
			fprintf(stderr,"Odczyt polecen ze standardowego wejscia\n");
			goto powtorz;
		}
		n = strlen(cmd);
		if (cmd[n-1] != '\n')
			fprintf(stderr,"Ostrzezenie: komenda zbyt dluga.\n");
		scmd = strtok(cmd, seplist);
		if (!scmd)
			goto powtorz;
		switch (scmd[0]) {
		case 't':
			/* Identyfikacja rodzaju terminala */
			if (dfn)
				fprintf(stderr,"Typ terminala juz zostal wybrany.\n");
			else {
				dfn = disp_init();
				/* Inicjalizacja mechanizmow prezentacji wynikow */
				if (dfn)
					dfn->init();
			}
			break;
		case 'p':	/* Pokazanie punktow na ekranie */
			if (dfn) {
				/* Obliczamy wspolrzedne punktow */
				przelicz(w, c, p, d);
				/* Pokazujemy punkty na ekranie */
				dfn->pokaz(w);
			}
			break;
		case 'k':	/* Koniec obliczen */
			break;
		case 'r':	/* Obrot ktoregos z wektorow wokol drugiego */
			par = strtok((char *) NULL, seplist);
			if (!par) {
				fprintf(stderr,"Rotacja punktow wymaga podania kata obrotu.\n");
				break;
			}
			a = atof(par);		/* Wartosc kata obrotu */
			a *= M_PI / 180.;	/* Przeliczenie na miare lukowa */
			/* Obrot moze nastapic wokol jednej z trzech osi */
			switch (scmd[1]) {
			case 'c':
				w1 = c;	/* Obrot wokol wektora wskazujacego srodek ekranu */
				w2 = pp;
				w3 = ph;
				break;
			case 'v':
				w1 = pp; /* Obrot wokol osi pionowej (zmiana kierunku patrzenia) */
				w2 = ph;
				w3 = c;
				break;
			case 'h':
				w1 = ph; /* Obrot wokol wektora poziomego (rownoleglego do ekranu - pochylanie aparatu) */
				w2 = c;
				w3 = pp;
				break;
			default:
				goto blad;
			}
			/* w1 - os obrotu; w2 i w3 - wektory obracane */
			macierz_obrotu(m, w1, a);
			mawek3(m, w2, t);
			dcopy3(t, w2);
			mawek3(m, w3, t);
			dcopy3(t, w3);
			/* Zawsze trzeba obrocic wektor wskazujacy pion na ekranie */
			mawek3(m, p, t);
			dcopy3(t, p);
			break;
		case 'd':	/* Odleglosc ekranu od obserwatora */
			d = dblcmd();
			break;
		case 's':	/* Zapisanie do pliku wynikow obliczen */
			par = strtok((char *) NULL, seplist);
			f=fopen(par,"w");
			if (!f) {
				printf("Blad otwarcia pliku %s\n",par);
				break;
			}
			fprintf(f,"# Wspolrzedne punktow na ekranie, d=%lf\n",d);
			fprintf(f,"# Wektor wyznaczajacy srodek ekranu: %.12lg %.12lg %.12lg\n",c[0],c[1],c[2]);
			fprintf(f,"# Wektor wyznaczajacy pion: %.12lg %.12lg %.12lg\n",p[0],p[1],p[2]);
			pokaz(f,w);
			fclose(f);
			break;
		case 'h':	/* Schowanie obiektow */
			vislist(w, -1);
			break;
		case 'u':	/* Pokazanie obiektow */
			vislist(w, 1);
			break;
		case 'l':
			/* Nazwa obiektu */
			par = strtok((char *) NULL, seplist);
			if (!par)
				break;
			wb = scan_list(w, par);
			if (wb) {
				/* przesuniecie opisu w pionie */
				wb->omin += dblcmd();
				/* i w poziomie */
				wb->hmin += dblcmd();
			}
			break;
		case 'w':	/* ustawienie jednego z wektorow */
			par = strtok((char *) NULL, seplist + 2);
			switch (scmd[1]) {
			case 'c':
				sscanf(par,"%lg %lg %lg",&c[0],&c[1],&c[2]);
				break;
			case 'p':
				sscanf(par,"%lg %lg %lg",&p[0],&p[1],&p[2]);
				break;
			default:
				goto blad;
			}
			/* nowe wektory pomocnicze: poziomu */
			axb3(c, p, ph);
			/* i pionu */
			axb3(ph, c, pp);
			break;
		case 'c':	/* Komenda bezposrednio do programu roboczego */
			if (dfn) {
				par = strtok((char *) NULL, seplist + 2);
				if (!par) {
					fprintf(stderr,"Nieoczekiwany koniec polecenia");
					break;
				}
				dfn->cmd(par);
				scmd[0] = 'c';
			}
			break;
		case 'i':	/* odczyt polecen z pliku */
			par = strtok((char *) NULL, seplist + 2);
			fi = fopen(par,"r");
			if (!fi) {
				perror("fopen");
				fi = stdin;
			} else
				fprintf(stderr,"Odczyt polecen z pliku %s\n",par);
			break;
		default:
		blad:
			printf("nieprawidlowe polecenie: %s\n",scmd);
		}
	} while (scmd[0] != 'k');
}

int main(int argc, char **argv) {
	double hmax, cen[3], pion[3], c;
	FILE *f;
	struct wektor *wlist, *wb, *wp;
	int i;
	setlinebuf(stdout);
	if (argc < 2)
		blad("Podaj nazwe pliku z danymi jako argument programu");
	f = fopen(argv[1], "r");
	if (!f) {
		perror("fopen");
		blad("Blad otwarcia pliku z danymi");
	}
	/* Alokacja pamieci na pierwszy wektor */
	wlist = (struct wektor *) xmalloc(sizeof(struct wektor));
	wb = wlist;
	/* Zerujemy wektor centrum */
	fill3(cen, 0.);
	/* Maksymalna wysokosc inicjujemy wartoscia tak mala, ze na pewno takiej nie bedzie */
	hmax = -1.e30;
	/* Opuszczamy linie komentarza */
	for (i = getc(f); i == (int) '#' || i == (int) '\n'; i = getc(f)) {
		putc(i, stdout);
		while (putc(getc(f), stdout) != (int) '\n');
	}
	ungetc(i, f);
	i = 0;
	/* Wczytanie kolejnej linii danych */
	while(fscanf(f,"%lg %lg %lg %lg %lg %lg %lg %s",&(wb->o),&(wb->hmin),&(wb->omin),wb->w,wb->w+1,wb->w+2,&(wb->h),wb->nazwa) == 8) {
		if (strlen(wb->nazwa) >= MAXNAZ)
			blad("Nazwa obieku jest za dluga.");
		/* Unormowanie wektora kierunkowego, zeby centrum bylo sensowne */
		c = adotb3(wb->w, wb->w);
		dscal3(wb->w, 1./sqrt(c));
		/* Dodajemy biezacy wektor do wektora centrum - bedzie usredniony */
		aplusb3(cen, wb->w, cen);
		/* Szukamy punktu, dla ktorego swiatlo bieglo najwyzej
		   - do niego bedzie skierowany wstepny pion */
		if (wb->hmin > hmax) {
			hmax = wb->hmin;
			dcopy3(wb->w, pion);
			printf("kandydat na wyznaczenie pionu: %s\n",wb->nazwa);
		}
		/* Ustawiamy znacznik widocznosci */
		wb->vis = 1;
		/* Element "omin" nie ma znaczenia w tym programie;
		   bedzie wykorzystywany do przesuwania podpisu w pionie. */
		wb->omin = 0.;
		/* Podobnie element "hmin", ktory bedzie uzywany
		   do przesuwania podpisu w poziomie. */
		wb->hmin = 0.;
		/* Alokacja pamieci na nastepny wektor */
		wb->n = (struct wektor *) xmalloc(sizeof(struct wektor));
		/* Zapamietujemy biezacy wskaznik */
		wp = wb;
		/* I nastepny na liscie staje sie biezacy */
		wb = wb->n;
		i++;
	}
	fclose(f);
	printf("Wczytano dane, liczba obserwowanych punktow wynosi %d.\n",i);
	if (i < 2)
		blad("Potrzebne sa przynajmniej 2 punkty."); 
	/* Ostatni element listy nie jest potrzebny - nie zostal wypelniony */
	free(wb);
	/* Znacznik konca listy */
	wp->n = (struct wektor *) NULL;
	/* Skalowanie wektora wyznaczajacego srodek ekranu */
	dscal3(cen, 1. / (double) i);
	/* Wywolanie interpretera */
	interp(wlist,cen,pion);
	/* Zamkniecie mechanizmow prezentacji wynikow */
	if (dfn)
		dfn->fin();
	return 0;
}
