/* Copyright (C) Krzysztof Strasburger, 2009
   licencja GNU GPL, wersja 2 lub pniejsza */

/* Operacje na macierzach i wektorach, do wlaczania do kodu */

/* Pomnozenie macierzy a i b, 3x3, przechowywanych wierszami. Wynik w c. */
static inline void mamul3(double *a, double *b, double *c) {
	c[0]=a[0]*b[0]+a[1]*b[3]+a[2]*b[6];
	c[1]=a[0]*b[1]+a[1]*b[4]+a[2]*b[7];
	c[2]=a[0]*b[2]+a[1]*b[5]+a[2]*b[8];
	c[3]=a[3]*b[0]+a[4]*b[3]+a[5]*b[6];
	c[4]=a[3]*b[1]+a[4]*b[4]+a[5]*b[7];
	c[5]=a[3]*b[2]+a[4]*b[5]+a[5]*b[8];
	c[6]=a[6]*b[0]+a[7]*b[3]+a[8]*b[6];
	c[7]=a[6]*b[1]+a[7]*b[4]+a[8]*b[7];
	c[8]=a[6]*b[2]+a[7]*b[5]+a[8]*b[8];
}

/* Pomnozenie macierzy 3x3 a przez wektor 3-elementowy b. Wynik w c */

static inline void mawek3(double *a, double *b, double *c) {
	c[0]=a[0]*b[0]+a[1]*b[1]+a[2]*b[2];
	c[1]=a[3]*b[0]+a[4]*b[1]+a[5]*b[2];
	c[2]=a[6]*b[0]+a[7]*b[1]+a[8]*b[2];
}

/* Iloczyn wektorowy wektorow a i b, wynik w c */
static inline void axb3(double *a, double *b, double *c) {
	c[0] = a[1] * b[2] - a[2] * b[1];
	c[1] = a[2] * b[0] - a[0] * b[2];
	c[2] = a[0] * b[1] - a[1] * b[0];
}

/* Iloczyn skalarny wektorow a i b */
static inline double adotb3(double *a, double *b) {
	return (a[0]*b[0] + a[1]*b[1] + a[2]*b[2]);
}

/* Suma wektorow a i b, wynik w c */
static inline void aplusb3(double *a, double *b, double *c) {
	c[0] = a[0] + b[0];
	c[1] = a[1] + b[1];
	c[2] = a[2] + b[2];
}

/* Suma wektorow a i x*b, wynik w c */
static inline void aplusxb3(double *a, double *b, double *c, double x) {
	c[0] = a[0] + x * b[0];
	c[1] = a[1] + x * b[1];
	c[2] = a[2] + x * b[2];
}

/* Roznica wektorow a i b, wynik w c */
static inline void aminusb3(double *a, double *b, double *c) {
	c[0] = a[0] - b[0];
	c[1] = a[1] - b[1];
	c[2] = a[2] - b[2];
}

/* Pomnozenie wektora a przez liczbe c */

static inline void dscal3(double *a, double c) {
	a[0] *= c;
	a[1] *= c;
	a[2] *= c;
}

/* Skopiowanie wektora a do b */
static inline void dcopy3(double *a, double *b) {
	b[0] = a[0];
	b[1] = a[1];
	b[2] = a[2];
}

/* Wypelnienie wszystkich elementow wektora ta sama wartoscia */
static inline void fill3(double *w, double v) {
	w[0] = v;
	w[1] = v;
	w[2] = v;
}
