/* Copyright (C) Krzysztof Strasburger, 2009
   licencja GNU GPL, wersja 2 lub pniejsza */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <math.h>
#include <ctype.h>
#include <string.h>
#include "mw3.h"

int debug = 0;
int verb = 1;
int geom = 1;
int skan = 0;

extern void g2k();
extern void k2g();
extern double relips();
extern void prop_prosto();
extern void prop_ugiecie();
extern void pinit();
extern void tinit();
extern void pelips();
extern void blad();
extern void ptwart();
extern void pset();
extern void pget();
extern void deltah_init();

static void uzycie() {
	printf("Program do przewidywania widocznosci odleglych obiektow.\n");
	printf("Oblicza bieg wiazki swiatla po prostej i modeluje ugiecie w atmosferze.\n");
	printf("Uruchomienie programu: widoki [-hgcds]\n");
	printf("h - wypisanie ten komunikat na terminalu i zakonczenie dzialania,\n");
	printf("g - rezygnacja z obliczen czysto geometrycznych (biegu swiatla po prostej),\n");
	printf("c - ciche obliczenia; wypisane zostaja tylko koncowe wyniki obliczen,\n");
	printf("d - wypisanie wielu komunikatow z obliczen; uzywane w razie podejrzen o bledy;\n");
	printf("s krok - tryb skanowania (z dodatnim krokiem!); pozwala wykryc wielokrotne obrazy tego samego obiektu.\n");
	printf("Dane sa czytane ze standardowego wejscia, wyniki wypisywane na standardowe wyjscie\n");
	printf("Autorem programu jest Krzysztof Strasburger.\n");
	printf("Program moze byc rozpowszechniany na zasadach licencji GPL, wersja 2.\n");
	exit(0);
}

/* Odczyt kata okreslajacego szerokosc (lim=90)
   lub dlugosc geograficzna (lim=180) */
static double czytsd(char *sym, int lim) {
	int i, n, limw;
	double s, w, f;
	char c[32];
	limw = 180;
	f = 1.;
	w = 0.;
	for (i = 0; i < 3; i++) { 	/* Czytamy po kolei stopnie, minuty i sekundy katowe */
		scanf("%s",c);
		if (strlen(c) > 31)
			blad("Za dlugi lancuch znakow reprezentujacy liczbe.");
		c[0] = (char) toupper((int) c[0]);
		if (c[0] == sym[0] || c[0] == sym[1])
			goto strona_swiata;
		s = atof(c);
		if (s < 0. || s > (double) limw)
			goto zle_wsp;
		w = w + f * s;
		/* Jesli byla kropka dziesietna, to konczymy */
		if (strstr(c,"."))
			break;
		limw = 60;
		f /= 60.;
	}
	/* Wczytanie  strony swiata. */
	scanf("%s",c);
	c[0] = (char) toupper((int) c[0]);
	if (c[0] != sym[0] && c[0] != sym[1])
		goto zle_wsp;
	if (w > (double) lim)
		goto zle_wsp;
strona_swiata:
	if (c[0] == sym[1])
		w = -w;
	/* Przeliczenie na miare lukowa kata */
	return w * M_PI / 180.;
zle_wsp:
	blad("Nieprawidlowe wspolrzedne geograficzne!");
}

/* Wczytanie wspolrzenych geograficznych i wysokosci punktu */
static void czytwg(double *sdh, char *nazwa, int nmax) {
	if (verb)
		printf("Podaj wspolrzedne geograficzne i wysokosc nad poziomem morza.\n");
	/* Szerokosc geograficzna */
	sdh[0] = czytsd("NS",90);
	/* Dlugosc geograficzna */
	sdh[1] = czytsd("EW",180);
	/* Wysokosc punktu nad poziomem morza */
	scanf("%lg",sdh+2);
	/* Nazwa */
	scanf("%s",nazwa);
	if (strlen(nazwa) >= nmax)
		blad("Za dluga nazwa obiektu.");
}

/* Wypisanie wspolrzednej geograficznej */
static void piszsd(double w, char *sym) {
	int i, d, m;
	double s;
	i = 0;
	if (w < 0.) {
		w = -w;
		i = 1;
	}
	/* Zamiana na miare katowa */
	w *= 180. / M_PI;
	printf("   %.4lf",w);
	d = floor(w);
	w = (w - d) * 60;
	m = floor(w);
	s = (w - m) * 60;
	printf(" (%3d %2d\'%4.1lf\") %c",d,m,s,sym[i]);
}

/* Wypisanie odleglosci punktu od punktu poczatkowego,
  jego wspolrzednych geograficznych i wysokosci nad poziomem morza. */
void piszp(double o, double *sdh) {
	/* odleglosc (w km; obliczenia sa prowadzone w metrach */
	printf("%.3lf",o/1000.);
	/* szerokosc geograficzna */
	piszsd(sdh[0], "NS");
	/* dlugosc geograficzna */
	piszsd(sdh[1], "EW");
	/* wysokosc */
	printf("   %.1lf\n",sdh[2]);
}

#define MAXNAZ	64	/* Maksymalna dlugosc nazwy obiektu */
int main (int argc, char **argv) {
	double xyz1[3],xyz2[3],rsdh1[4],rsdh2[4],w[6],s,a,b;
	double hrsd[4],pt[4],h2,dh,htmp,ptmp;
	int n, i, j, jest, dalej, len;
	char znak, naz1[MAXNAZ],naz2[MAXNAZ];
#ifndef WINDOWS
	/* Buforowanie standardowego wyjscia - jedna linia */
	setlinebuf(stdout);
#endif
	/* Tu bedzie interpretacja argumentow */
	for (i = 1; i < argc; i++) {
		argv++;	/* W zerowym argumencie jest nazwa programu, dlatego zwiekszamy argv na poczatku petli. */ 
		len = strlen(*argv);
		for (j = 0; j < len; j++) {
			switch ((*argv)[j]) {
			case '-':
				break;
			case 'd':
				debug = 1;
				break;
			case 'c':
				verb = 0;
				break;
			case 'g':
				geom = 0;
				break;
			case 's':
				if (j+1 != len)
					uzycie();
				argv++;
				skan = atoi(*argv);
				printf("# Skanowanie z parametrem wysokosci %d m\n",skan);
				if (skan < 1)
					uzycie();
				i++;
				break;
			case 'h':
				uzycie();
			default:
				printf("Nieznany argument %c\n",(*argv)[j]);
				uzycie();
			}
		}
	}
	jest = 0;
	s = 1000.;
	/* Petla interpretera */
	/* czytamy standardowe wejscie az do napotkania znaku roznego od spacji
	   lub tabulatora; opuszczamy tez puste linie */
	do {
		do
			if (scanf("%c",&znak) == EOF)
				blad("Osiagnieto koniec pliku.");
		while (znak == ' ' || znak == '\t' || znak == '\n' || znak == '\r');
		znak = (char) tolower((int) znak);
		switch (znak) {
		case 'k':	/* Koniec programu */
			break;
		case '#':	/* linia komentarza */
			do
				if (scanf("%c",&znak) == EOF)
					blad("Osiagnieto koniec pliku.");
			while (znak != '\n');
			break;
		case 'e':	/* Parametry elipsoidy */
			if (verb)
				printf("Podaj polosie rownikowa i biegunowa elipsoidy.\n");
			scanf("%lg %lg",&a,&b);
			pelips(a,b);
			break;
		case 'p':
			if (verb)
				printf("Parametry cisnienia\n");
			pinit();
			break;
		case 't':
			if (verb)
				printf("Parametry temperatury\n");
			tinit();
			break;
		case 'w':	/* Zaleznosc cisnienia i temperatury od wysokosci */
			/* Szerokosc geograficzna */
			hrsd[2] = czytsd("NS",90);
			/* Dlugosc geograficzna */
			hrsd[3] = czytsd("EW",180);
			/* Poczatkowa i koncowa wysokosc punktu nad poziomem morza oraz skok wysokosci */
			scanf("%lg %lg %lg",hrsd,&h2,&dh);
			if (h2 < hrsd[0] || dh <= 0.) {
				fprintf(stderr,"Nieprawidlowe parametry do obliczen cisnienia i temperatury.\n");
				break;
			} else {
				printf("   h      p      dp/dh      T      dT/dh\n");
				pget(&htmp, &ptmp);
			}
			do {
				ptwart(hrsd, pt);
				printf("%.2lf  %.1lf %.1lf %.4lf %.4lf\n",hrsd[0],pt[0],pt[1],pt[2]-273.15,pt[3]);
				hrsd[0] += dh;
			} while (hrsd[0] <= h2);
			pset(htmp, ptmp);
			break;
		case 's':
			if (verb)
				printf("Dlugosc pojedynczego kroku symulacji (w m): ");
			scanf("%lg",&s);
			if (s < 0.)
				blad("Dlugosc kroku musi byc dodatnia!");
			break;
		case 'h':	/* Przesuwanie profilu temperatury */
			if (!jest)
				blad("przed przesunieciem profilu temperatury nalezy podac punkt obserwacji");
			if (verb)
				printf("Przesuniecie profilu temperatury;\npodaj parametry rozkladu:");
			scanf("%lg %lg %lg %d",&b,&a,&dh,&i);
			deltah_init(xyz1,b,a,dh,i);
			break;
		case 'm':
			if (verb)
				printf("Miejsce w terenie\n");
			if (!jest) {
				if (verb)
					printf("Podaj miejsce obserwacji.\n");
				czytwg(rsdh1+1,naz1,MAXNAZ);	/* Wspolrzedne geograficzne */
				rsdh1[0] = relips(rsdh1[1]); /* "poziom morza" na elipsoidzie */
				rsdh1[0] += rsdh1[3];	/* dodanie wysokosci */
				g2k(rsdh1, xyz1);	/* Obliczenie wspolrzednych kartezjanskich */
				jest = 1;
				printf("# Widok z obiektu %s, o wysokosci %.1lfm\n",naz1,rsdh1[3]);
			} else {
				if (verb)
					printf("Podaj obserwowany punkt.\n");
				/* Czynnosci jak wyzej, ale dla drugiego punktu */
				czytwg(rsdh2+1,naz2,MAXNAZ);
				rsdh2[0] = relips(rsdh2[1]);
				rsdh2[0] += rsdh2[3];
				g2k(rsdh2, xyz2);
				if (geom) {
					if (verb)
						printf("Rozchodzenie swiatla po prostej\n");
					prop_prosto(xyz1, xyz2, s, &w[0]);
					printf("%.3lf %.3lf %.3lf %.12lg %.12lg %.12lg %.1lf %s\n",w[0]*0.001,w[1],w[2]*0.001,w[3],w[4],w[5],rsdh2[3],naz2);
				}
				if (verb)
					printf("Symulacja ugiecia swiatla w atmosferze\n");
				dalej = 0;
powtorz:
				prop_ugiecie(xyz1, xyz2, s, &w[0], skan, &dalej, rsdh2[3]);
				if (dalej) {
					printf("%.3lf %.3lf %.3lf %.12lg %.12lg %.12lg %.1lf %s\n",w[0]*0.001,w[1],w[2]*0.001,w[3],w[4],w[5],rsdh2[3],naz2);
					if (skan) goto powtorz;
				}
			}
			break;
		default:
			fprintf(stderr,"Wczytano znak %c. ",znak);
			blad("Nieznane polecenie.");
		}
	} while (znak != 'k');
	return 0;
}
