/* Copyright (C) Krzysztof Strasburger, 2009
   licencja GNU GPL, wersja 2 lub pniejsza */

#include <math.h>

/* Operacje na wspolrzednych kartezjanskich i geograficznych na elipsoidzie */

extern void blad();

/* Polosie rownikowa i biegunowa elipsoidy obrotowej */
/* Domyslne wartosci opisuja elipsoide Krasowskiego */
#define AKRASOW 6378245.
#define BKRASOW 6356863.019
double a = AKRASOW;
double b = BKRASOW;
double aa = AKRASOW * AKRASOW;
double bb = BKRASOW * BKRASOW;
double cc = AKRASOW * AKRASOW - BKRASOW * BKRASOW;

/* Ustawienie wlasnych parametrow elipsoidy */
void pelips(double ax, double bx) {
	a = ax;
	b = bx;
	if (a <= 0. || b <= 0)
		blad("Polosie elipsoidy musza byc dodatnie.");
	if (a < b)
		blad("Polos rownikowa nie moze byc krotsza od biegunowej.");
	aa = a * a;
	bb = b * b;
	cc = aa - bb;
}

/* Przeliczenie wspolrzednych biegunowych na kartezjanskie */
static inline void b2k(double r, double t, double *x, double *y) {
	*x = r * cos(t);
	*y = r * sin(t);
}

/* Przeliczenie wspolrzednych geograficznych: promien, szerokosc
   i dlugosc geograficzna, na kartezjanskie: x, y, z;
   os z przechodzi przez bieguny, os x przez poludniki 0 i 180. */
void g2k(double *rtf, double *xyz) {
	double xy;
	b2k(rtf[0], rtf[1], &xy, xyz+2);
	b2k(xy, rtf[2], xyz, xyz+1);
}

/* Przeliczenie wspolrzednych kartezjanskich na biegunowe;
   czesc operacji na wspolrzednych geograficznych, wiec x jest zawsze
   dodatnie. */
static inline void k2b(double x, double y, double *r, double *t) {
	*r = sqrt(x*x+y*y);
	*t = 0.;
	if (*r > 0)
		*t = asin (y / *r);
}

/* Przeliczenie wspolrzednych kartezjanskich na geograficzne. */
void k2g(double *xyz, double *rtf) {
	double rxy, df;
	/* Rzut wektora na plaszczyzne xy ("rownika") */
	rxy = sqrt(xyz[0] * xyz[0] + xyz[1] * xyz[1]);
	/* Obliczenie promienia i szerokosci geograficznej. */
	k2b(rxy, xyz[2], rtf, rtf+1);
	/* Obliczenie dlugosci geograficznej. */
	rtf[2] = 0.;
	if (rxy > 0.)
	rtf[2] = asin(xyz[1] / rxy);
	if (xyz[0] < 0.) {
		df = M_PI_2;
		if (xyz[1] < 0.)
			df = -df;
		rtf[2] += df;
	}
}

/* Obliczenie odleglosci punktu na elipsoidzie obrotowej od poczatku
   ukladu wspolrzednych; dana jest szerokosc geograficzna t */
double relips(double t) {
	double s, c;
	s = sin(t) / b;
	c = cos(t) / a;
	return (1./sqrt(s*s + c*c));
}

/* Znalezienie lokalnego pionu, prostopadlego do powierzchni elipsoidy */
/* x - wspolrzedne kartezjanskie (dane)
   rsd - odleglosc od srodka elipsoidy i wspolrzedne geograficzne,
         obliczane "po drodze" (wynik)
   p - wektor kierunkowy lokalnego pionu (wynik)
   eps - dopuszczalne odchylenie przy podstawie (dane)
   Funkcja zwraca wysokosc nad poziomem morza.

   Obliczenia sa iteracyjne, poniewaz analityczne wzory wymagalyby
   rozwiazania rownania 4-go stopnia.
*/

double pion(double *x, double *rsd, double *p, double eps) {

	double r, rg, xl, yl, a2x, b2y, dx, dy, d, xy[4], rf[2];
	double *x1, *x2, *tmp;

/* Obliczenie wspolrzednych geograficznych (od razu mamy wspolrzedne
   biegunowe punktu w lokalnym, dwuwymiarowym ukladzie wspolrzednych) */
	k2g(x, rsd);
/* Znajdujemy punkt na elipsie (i na wektorze ze srodka elipsy) */
	r = relips(rsd[1]);
/* "wysokosc" w kierunku ze srodka elipsoidy */
	rg = rsd[0] - r;
	/* if (pion_centr) {
		dcopy3(x, p)
		return rg;
	} */

/* Podzial pamieci roboczej */
	x1=&xy[0]; x2=x1+2;

/* Obliczenie wspolrzednych kartezjanskich w lokalnym, dwuwymiarowym
   ukladzie wspolrzednych (dla danej dlugosci geograficznej) */
	xl = sqrt(x[0]*x[0]+x[1]*x[1]);
	yl = x[2];
/* Stale skladniki obliczen */
	a2x = aa * xl;	/* a^2*x */
	b2y = bb * yl;	/* b^2*y */
/* Kopia wspolrzednych do obliczen */
	x1[0] = xl;
	x1[1] = yl;
	b2k(r, rsd[1], x2, x2+1);
powt:
/* Zamiana punktow; w x1 otrzymamy nowy punkt na elipsie, w x2 - punkt stary
   (w pierwszym przebiegu jest to po prostu punkt, w ktoeym szukamy pionu */
	tmp=x1; x1=x2; x2=tmp;
/* Obliczenie odleglosci miedzy tymi punktami. */
	dx = x2[0] - x1[0];
	dy = x2[1] - x1[1];
	d = sqrt(dx*dx+dy*dy);
	if (d > eps) {
/* Obliczenie kolejnego przyblizenia punktu na elipsie; najpierw przesuwamy
   stary punkt tak, by znalazl sie blizej lokalnego pionu; zadamy, by wektor
   prowadzacy z punktu szukanego do rozpatrywanego, byl prostopadly do poziomu
   w ostatnio znalezionym przyblizonym punkcie na elipsie. */
		if (a2x > b2y) {
			x2[0] = x1[0];
			x2[1] = (b2y + cc * x1[1]) * x1[0] / a2x;
		} else {
			x2[0] = (a2x - cc * x1[0]) * x1[1] / b2y;
			x2[1] = x1[1];
		}
/* Ten punkt (zwykle poza elipsa) rzutujemy na elipse w kierunku do jej srodka. */
		k2b(x2[0], x2[1], rf, rf+1);
		r = relips(rf[1]);
		b2k(r, rf[1], x2, x2+1);
/* i to jest kolejne przyblizenie*/
		goto powt;
	}
/* W tym miejscu w x1 jest punkt na elipsie, z ktorego nalezy poprowadzic 
   lokalny pion. Skladowe wektora kierunkowego bierzemy z rownania
   normalnej do elipsy. Skladowa z-owa jest rowna lokalnej y-owej/ */
	p[2] = x1[1] * aa / bb; /* a tu cos jest zle */
/* Teraz wykorzystujemy dlugosc geograficzna
   do obliczenia skladowych x i y wektora pionu */
	b2k(x1[0], rsd[2], p, p+1);
/* i obliczamy wysokosc nad poziomem morza */
	dx = xl - x1[0];
	dy = yl - x1[1];
	r = sqrt(dx * dx + dy * dy);
	if (rg < 0.)
		r = -r;
	return r;
}
